package minischeme

/**
 * Representation for a piece of code. Also tracks register usage and
 * activation frame size.
 *
 * @author Michel Schinz
 */

class Code() {
  // Frame size
  private var frameSize: Int = 0
  private var maxFrameSize: Int = 0

  def getFrameSize: Int = frameSize
  def incFrameSize(delta: Int): Unit = {
    frameSize += delta
    maxFrameSize = maxFrameSize max frameSize
  }
  def decFrameSize(delta: Int): Unit =
    frameSize -= delta

  def getMaxFrameSize: Int = maxFrameSize

  // Registers
  private var freeRegs: List[Register] = Registers.allocatable

  def allocateRegister(): Register = freeRegs match {
    case r :: rest => freeRegs = rest; r
    case _ => error("No more free registers")
  }
  def allocateRegister(r: Register): Register =
    if (freeRegs contains r) {
      freeRegs = freeRegs -- List(r)
      r
    } else
      error("Register " + r + " already allocated")

  def allocatedRegisters: List[Register] =
    Registers.allocatable -- freeRegs

  def freeRegister(r: Register) =
    freeRegs = r :: freeRegs

  def withFreshRegister[T](body: Register=>T): T = {
    val r = allocateRegister()
    try {
      body(r)
    } finally {
      freeRegister(r)
    }
  }

  // Instructions
  private val items = new scala.collection.mutable.ListBuffer[Representable]()

  def emit(i: Instruction): Unit =
    items += i
  def emit(op: OpcodeRRR, r1: Register, r2: Register, r3: Register): Unit =
    emit(new InstructionRRR(op, r1, r2, r3))
  def emit(op: OpcodeRRR, r1: Register, r2: Register, c: Int): Unit =
    if (c == 0)
      emit(op, r1, r2, Registers.R0)
    else withFreshRegister { reg =>
      emit(LINT, reg, c)
      emit(op, r1, r2, reg)
    }
  def emit(op: ALOC.type, r1: Register, r2: Register): Unit =
    emit(new InstructionRR(op, r1, r2))
  def emit(op: LINT.type, r: Register, i: Int): Unit =
    emit(new InstructionRC(op, r, i))
  def emit(op: LINT.type, r: Register, l: Label): Unit =
    emit(new InstructionRC(op, r, l))
  def emit(op: OpcodeR, r: Register): Unit =
    emit(new InstructionR(op, r))
  def emit(op: HALT.type): Unit =
    emit(new InstructionHalt(op))
  def emit(label: Label): Unit =
    items += new Representable() {
      override def representation: String = label.representation + ":"
    }

  def addComment(comment: String): Unit =
    items += new Representable() {
      override def representation: String = "; " + comment
    }

  def emitPush(reg: Register): Unit = {
    emit(STOR, reg, Registers.FP, getFrameSize)
    incFrameSize(1)
  }
  def emitPop(reg: Register): Unit = {
    decFrameSize(1)
    emit(LOAD, reg, Registers.FP, getFrameSize)
  }

  // Dumping
  def dumpTo(stream: java.io.PrintStream): Unit = {
    for (i <- items)
      stream.println(i.representation)
  }
}
