package minischeme

import java.io.PrintStream
import scala.collection.immutable.{Set,ListSet,Map,ListMap}
import SymbolicTreeModule._

/**
 * Code generator for minischeme.
 *
 * @author Michel Schinz
 */
object Generator extends ((List[Tree],PrintStream) => Unit) {
  private val auxCodes = new scala.collection.mutable.ListBuffer[Code]()
  private type Env = PartialFunction[Symbol, Int]

  private val compPrim = ListSet(Primitives.lt, Primitives.le, Primitives.eq,
                                 Primitives.ge, Primitives.gt)
  private val arithPrim = ListSet(Primitives.add, Primitives.sub,
                                  Primitives.mul, Primitives.div,
                                  Primitives.mod)

  def apply(trees: List[Tree], stream: PrintStream): Unit = {
    // Attribute locations to global symbols
    var globalOffset: Int = 0
    trees foreach {
      case Define(sym, _) =>
        sym.location = GlobalLocation(globalOffset)
        globalOffset = globalOffset + 1
      case _ =>
    }

    // Allocate space for global variables
    val mainCode = new Code()
    mainCode.emit(LINT, Registers.GP, globalOffset)
    mainCode.emit(ALOC, Registers.GP, Registers.GP)

    // Allocate global "stack frame"
    val loadGlobalFrameSizeInstr = new InstructionRC(LINT, Registers.FP, 0)
    mainCode.emit(loadGlobalFrameSizeInstr)
    mainCode.emit(ALOC, Registers.FP, Registers.FP)

    // Generate program code
    trees foreach { gen(_, mainCode) }
    loadGlobalFrameSizeInstr.c = mainCode.getMaxFrameSize
    mainCode.emit(HALT)

    // Dump all code, starting with main one
    mainCode dumpTo stream
    auxCodes foreach (_ dumpTo stream)
  }

  private def gen(tree: Tree, code: Code): Unit = tree match {
    case Define(sym, value) =>
      code.addComment("begin " + sym.name)
      code withFreshRegister { tmpReg =>
        genLoad(value, tmpReg, code)
        sym.location match {
          case GlobalLocation(o) =>
            code.emit(STOR, tmpReg, Registers.GP, o)
        }
      }
      code.addComment("end " + sym.name)

    case _ =>
      code withFreshRegister { tmpReg => genLoad(tree, tmpReg, code) }
  }

  private def genCond(tree: Tree, when: Boolean, targetLabel: Label, code: Code): Unit =
  tree match {
    case And(c1, c2) if when =>
      val afterLabel = Label.newLabel("after")
      genCond(c1, false, afterLabel, code)
      genCond(c2, true, targetLabel, code)
      code.emit(afterLabel)
    case And(c1, c2) if !when =>
      genCond(c1, false, targetLabel, code)
      genCond(c2, false, targetLabel, code)

    case Or(c1, c2) =>
      val and = And(Prim(Primitives.not, List(c1)),
                    Prim(Primitives.not, List(c2)))
      genCond(and, !when, targetLabel, code)

    case Prim(sym, List(e1, e2)) if compPrim(sym) =>
      code withFreshRegister { r1 =>
        genLoad(e1, r1, code)
        code withFreshRegister { r2 =>
          genLoad(e2, r2, code)
          code withFreshRegister { targetReg =>
            code.emit(LINT, targetReg, targetLabel)
            sym match {
              case Primitives.lt =>
                code.emit(if (when) JLT else JGE, targetReg, r1, r2)
              case Primitives.le =>
                code.emit(if (when) JLE else JGT, targetReg, r1, r2)
              case Primitives.eq =>
                code.emit(if (when) JEQ else JNE, targetReg, r1, r2)
              case Primitives.ge =>
                code.emit(if (when) JGE else JLT, targetReg, r1, r2)
              case Primitives.gt =>
                code.emit(if (when) JGT else JLE, targetReg, r1, r2)
            }
          }
        }
      }

    case Prim(Primitives.not, expr :: Nil) =>
      genCond(expr, !when, targetLabel, code)

    case expr =>
      code withFreshRegister { tmpReg =>
        genLoad(expr, tmpReg, code)
        code withFreshRegister { targetReg =>
          code.emit(LINT, targetReg, targetLabel)
          code.emit(if (when) JNE else JEQ, targetReg, tmpReg, Registers.R0)
        }
      }
  }

  private def genLoad(tree: Tree, targetReg: Register, code: Code): Unit =
  tree match {
    case Lambda(args, body) =>
      // Code for lambda: load label
      val label = Label.newLabel("lambda")
      code.emit(LINT, targetReg, label)

      // Code for lambda body
      val lCode = new Code()
      lCode.emit(label)
      for (argReg <- 1 to args.length)
        lCode.allocateRegister(Registers(argReg))
      // allocate and link frame
      var loadFrameSizeInstr: InstructionRC = null
      lCode withFreshRegister { tmpReg =>
        loadFrameSizeInstr = new InstructionRC(LINT, tmpReg, 0)
        lCode.emit(loadFrameSizeInstr)
        lCode.emit(ALOC, tmpReg, tmpReg)
        lCode.emit(STOR, Registers.FP, tmpReg, Registers.R0)
        lCode.incFrameSize(1)
        lCode.emit(ADD, Registers.FP, tmpReg, Registers.R0)
      }
      // save all parameters (including LK)
      lCode.emitPush(Registers.LK)
      for ((sym, regIdx) <- args.zipWithIndex) {
        sym.location = FrameLocation(lCode.getFrameSize)
        lCode.emitPush(Registers(regIdx + 1))
        lCode.freeRegister(Registers(regIdx + 1))
      }
      val retValReg = lCode.allocateRegister(Registers(1))
      genLoad(body, retValReg, lCode)
      lCode.decFrameSize(args.length)
      lCode.emitPop(Registers.LK)
      lCode.emitPop(Registers.FP)
      assert(lCode.getFrameSize == 0, "non-zero frame size")
      lCode.emit(JEQ, Registers.LK, Registers.R0, Registers.R0) // return
      lCode.freeRegister(retValReg)
      loadFrameSizeInstr.c = lCode.getMaxFrameSize
      auxCodes += lCode

    case Let(sym, value, body) =>
      genLoad(value, targetReg, code)
      sym.location = FrameLocation(code.getFrameSize)
      code.emitPush(targetReg)
      genLoad(body, targetReg, code)
      code.decFrameSize(1)

    case Begin(body) =>
      for (statement <- body)
        genLoad(statement, targetReg, code)

    case If(cond, thenp, elsep) =>
      val elseLabel = Label.newLabel("else")
      genCond(cond, false, elseLabel, code)
      genLoad(thenp, targetReg, code)
      val afterLabel = Label.newLabel
      code withFreshRegister { tmpReg =>
        code.emit(LINT, tmpReg, afterLabel)
        code.emit(JEQ, tmpReg, Registers.R0, Registers.R0)
      }
      code.emit(elseLabel)
      genLoad(elsep, targetReg, code)
      code.emit(afterLabel)

    case And(c1, c2) =>
      genLoad(c1, targetReg, code)
      val afterLabel = Label.newLabel
      code withFreshRegister { tmpReg =>
        code.emit(LINT, tmpReg, afterLabel)
        code.emit(JEQ, tmpReg, targetReg, Registers.R0)
      }
      genLoad(c2, targetReg, code)
      code.emit(afterLabel)

    case Or(c1, c2) =>
      genLoad(c1, targetReg, code)
      val afterLabel = Label.newLabel
      code withFreshRegister { tmpReg =>
        code.emit(LINT, tmpReg, afterLabel)
        code.emit(JNE, tmpReg, targetReg, Registers.R0)
      }
      genLoad(c2, targetReg, code)
      code.emit(afterLabel)

    case Apply(function, args) =>
      // save registers
      code.freeRegister(targetReg)
      val registersToSave = code.allocatedRegisters
      for (r <- registersToSave) {
        code.emitPush(r)
        code.freeRegister(r)
      }
      val funReg = code.allocateRegister(Registers(28))
      genLoad(function, funReg, code)
      for (i <- 1 to args.length) {
        code.allocateRegister(Registers(i))
        genLoad(args(i - 1), Registers(i), code)
      }
      val retLabel = Label.newLabel("ret")
      code.emit(LINT, Registers.LK, retLabel)
      code.emit(JEQ, funReg, Registers.R0, Registers.R0)
      code.emit(retLabel)
      code.freeRegister(funReg)
      for (i <- 1 to args.length)
        code.freeRegister(Registers(i))
      val retValReg = Registers(1)
      code.allocateRegister(targetReg)
      if (targetReg != retValReg)
        code.emit(ADD, targetReg, retValReg, Registers.R0)
      // restore registers
      for (r <- registersToSave.reverse) {
        code.allocateRegister(r)
        code.emitPop(r)
      }

    case Prim(Primitives.vector, elems) =>
      code.emit(LINT, targetReg, elems.length)
      code.emit(ALOC, targetReg, targetReg)
      code withFreshRegister { eReg =>
        var offset = 0
        for (e <- elems) {
          genLoad(e, eReg, code)
          code.emit(STOR, eReg, targetReg, offset)
          offset = offset + 1
        }
      }

    case Prim(Primitives.vector_ref, List(vect, Num(0))) =>
      genLoad(vect, targetReg, code)
      code.emit(LOAD, targetReg, targetReg, Registers.R0)

    case Prim(Primitives.vector_ref, List(vect, index)) =>
      genLoad(vect, targetReg, code)
      code withFreshRegister { indexReg =>
        genLoad(index, indexReg, code)
        code.emit(LOAD, targetReg, targetReg, indexReg)
      }

    case Prim(Primitives.vector_set, List(vect, index, value)) =>
      genLoad(vect, targetReg, code)
      code withFreshRegister { indexReg =>
        genLoad(index, indexReg, code)
        code.withFreshRegister { valueReg =>
          genLoad(value, valueReg, code)
          code.emit(STOR, valueReg, targetReg, indexReg)
        }
      }

    case Prim(Primitives.print_char, List(c)) =>
      genLoad(c, targetReg, code)
      code.emit(PCHR, targetReg)
    case Prim(Primitives.read_char, List()) =>
      code.emit(RCHR, targetReg)

    case Prim(sym, List(e1, e2)) if arithPrim(sym) =>
      genLoad(e1, targetReg, code)
      code withFreshRegister { e2Reg =>
        genLoad(e2, e2Reg, code)
        sym match {
          case Primitives.add => code.emit(ADD, targetReg, targetReg, e2Reg)
          case Primitives.sub => code.emit(SUB, targetReg, targetReg, e2Reg)
          case Primitives.mul => code.emit(MUL, targetReg, targetReg, e2Reg)
          case Primitives.div => code.emit(DIV, targetReg, targetReg, e2Reg)
          case Primitives.mod => code.emit(MOD, targetReg, targetReg, e2Reg)
        }
      }

    case Ident(ident) =>
      ident.location match {
        case FrameLocation(index) =>
          code.emit(LOAD, targetReg, Registers.FP, index)
        case GlobalLocation(index) =>
          code.emit(LOAD, targetReg, Registers.GP, index)
      }

    case Num(value) =>
      code.emit(LINT, targetReg, value)

    case expr =>
      genLoad(If(expr, Num(1), Num(0)), targetReg, code)
  }
}
