#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

#include "opcode.h"
#include "fail.h"
#include "memory.h"
#include "engine.h"

char* engine_identity = "threaded";

typedef void* instrptr_t;

typedef struct {
  instrptr_t code;
  instrptr_t next[0];
} instr_t;

typedef struct {
  instrptr_t code;
  reg_idx_t r;
  instrptr_t next[0];
} instr_r_t;

typedef struct {
  instrptr_t code;
  reg_idx_t ra, rb;
  instrptr_t next[0];
} instr_rr_t;

typedef struct {
  instrptr_t code;
  reg_idx_t r;
  intptr_t c;
  instrptr_t next[0];
} instr_rc_t;

typedef struct {
  instrptr_t code;
  reg_idx_t ra, rb, rc;
  instrptr_t next[0];
} instr_rrr_t;

typedef enum {
  engine_mode_init,
  engine_mode_run
} engine_mode_t;

static void* threading_labels[OPCODE_COUNT];
static void* curr_addr;
static intptr_t regs[REGS_COUNT];

static int init_or_run(engine_mode_t engine_mode);
static void check_available_memory(size_t size);

void engine_setup() {
  curr_addr = memory_get_start();
  init_or_run(engine_mode_init);

  for (unsigned r = 0; r < REGS_COUNT; ++r)
    regs[r] = 0;
}

void engine_cleanup() {
  // nothing to do
}

void engine_emit(opcode_t opcode) {
  check_available_memory(sizeof(instr_t));
  instr_t* instr = curr_addr;
  instr->code = threading_labels[opcode];
  curr_addr = instr->next;
}

void engine_emit_r(opcode_t opcode, reg_idx_t r) {
  check_available_memory(sizeof(instr_r_t));
  instr_r_t* instr = curr_addr;
  instr->code = threading_labels[opcode];
  instr->r = r;
  curr_addr = instr->next;
}

void engine_emit_rr(opcode_t opcode, reg_idx_t ra, reg_idx_t rb) {
  check_available_memory(sizeof(instr_rr_t));
  instr_rr_t* instr = curr_addr;
  instr->code = threading_labels[opcode];
  instr->ra = ra;
  instr->rb = rb;
  curr_addr = instr->next;
}

void engine_emit_rc(opcode_t opcode, reg_idx_t r, intptr_t c) {
  check_available_memory(sizeof(instr_rc_t));
  instr_rc_t* instr = curr_addr;
  instr->code = threading_labels[opcode];
  instr->r = r;
  instr->c = c;
  curr_addr = instr->next;
}

void engine_emit_rrr(opcode_t opcode,
                     reg_idx_t ra,
                     reg_idx_t rb,
                     reg_idx_t rc) {
  check_available_memory(sizeof(instr_rrr_t));
  instr_rrr_t* instr = curr_addr;
  instr->code = threading_labels[opcode];
  instr->ra = ra;
  instr->rb = rb;
  instr->rc = rc;
  curr_addr = instr->next;
}

void* engine_get_next_address() {
  return curr_addr;
}

void engine_patch(void* instr_to_patch, void* target) {
  instr_rc_t* instr = instr_to_patch;
  instr->c = (intptr_t)target;
}

int engine_run() {
  return init_or_run(engine_mode_run);
}

intptr_t engine_get_register(reg_idx_t r) {
  return regs[r];
}

static int init_or_run(engine_mode_t engine_mode) {
  if (engine_mode == engine_mode_init) {
    // Initialize the "global" array of instruction pointers, to be
    // used by the emitting methods.
    threading_labels[opcode_ADD] = &&l_ADD;
    threading_labels[opcode_SUB] = &&l_SUB;
    threading_labels[opcode_MUL] = &&l_MUL;
    threading_labels[opcode_DIV] = &&l_DIV;
    threading_labels[opcode_MOD] = &&l_MOD;
    threading_labels[opcode_JLT] = &&l_JLT;
    threading_labels[opcode_JLE] = &&l_JLE;
    threading_labels[opcode_JEQ] = &&l_JEQ;
    threading_labels[opcode_JNE] = &&l_JNE;
    threading_labels[opcode_JGE] = &&l_JGE;
    threading_labels[opcode_JGT] = &&l_JGT;
    threading_labels[opcode_LINT] = &&l_LINT;
    threading_labels[opcode_LOAD] = &&l_LOAD;
    threading_labels[opcode_STOR] = &&l_STOR;
    threading_labels[opcode_ALOC] = &&l_ALOC;
    threading_labels[opcode_RCHR] = &&l_RCHR;
    threading_labels[opcode_PCHR] = &&l_PCHR;
    threading_labels[opcode_HALT] = &&l_HALT;
    return 0;
  }

  instrptr_t* pc = memory_get_start();
  goto **pc;

 l_ADD: {
    instr_rrr_t* instr = (instr_rrr_t*)pc;
    regs[instr->ra] = regs[instr->rb] + regs[instr->rc];
    pc = instr->next;
    goto **pc;
  }

 l_SUB: {
    instr_rrr_t* instr = (instr_rrr_t*)pc;
    regs[instr->ra] = regs[instr->rb] - regs[instr->rc];
    pc = instr->next;
    goto **pc;
  }

 l_MUL: {
    instr_rrr_t* instr = (instr_rrr_t*)pc;
    regs[instr->ra] = regs[instr->rb] * regs[instr->rc];
    pc = instr->next;
    goto **pc;
  }

 l_DIV: {
    instr_rrr_t* instr = (instr_rrr_t*)pc;
    if (regs[instr->rc] == 0)
      fail("division by zero");
    regs[instr->ra] = regs[instr->rb] / regs[instr->rc];
    pc = instr->next;
    goto **pc;
  }

 l_MOD: {
    instr_rrr_t* instr = (instr_rrr_t*)pc;
    if (regs[instr->rc] == 0)
      fail("division by zero");
    regs[instr->ra] = regs[instr->rb] % regs[instr->rc];
    pc = instr->next;
    goto **pc;
  }

 l_JLT: {
    instr_rrr_t* instr = (instr_rrr_t*)pc;
    pc = (regs[instr->rb] < regs[instr->rc])
      ? (instrptr_t*)regs[instr->ra]
      : instr->next;
    goto **pc;
  }

 l_JLE: {
    instr_rrr_t* instr = (instr_rrr_t*)pc;
    pc = (regs[instr->rb] <= regs[instr->rc])
      ? (instrptr_t*)regs[instr->ra]
      : instr->next;
    goto **pc;
  }

 l_JEQ: {
    instr_rrr_t* instr = (instr_rrr_t*)pc;
    pc = (regs[instr->rb] == regs[instr->rc])
      ? (instrptr_t*)regs[instr->ra]
      : instr->next;
    goto **pc;
  }

 l_JNE: {
    instr_rrr_t* instr = (instr_rrr_t*)pc;
    pc = (regs[instr->rb] != regs[instr->rc])
      ? (instrptr_t*)regs[instr->ra]
      : instr->next;
    goto **pc;
  }

 l_JGE: {
    instr_rrr_t* instr = (instr_rrr_t*)pc;
    pc = (regs[instr->rb] >= regs[instr->rc])
      ? (instrptr_t*)regs[instr->ra]
      : instr->next;
    goto **pc;
  }

 l_JGT: {
    instr_rrr_t* instr = (instr_rrr_t*)pc;
    pc = (regs[instr->rb] > regs[instr->rc])
      ? (instrptr_t*)regs[instr->ra]
      : instr->next;
    goto **pc;
  }

 l_LINT: {
    instr_rc_t* instr = (instr_rc_t*)pc;
    regs[instr->r] = instr->c;
    pc = instr->next;
    goto **pc;
  }

 l_LOAD: {
    instr_rrr_t* instr = (instr_rrr_t*)pc;
    regs[instr->ra] = ((intptr_t*)regs[instr->rb])[(int)regs[instr->rc]];
    pc = instr->next;
    goto **pc;
  }

 l_STOR: {
    instr_rrr_t* instr = (instr_rrr_t*)pc;
    ((intptr_t*)regs[instr->rb])[(int)regs[instr->rc]] = regs[instr->ra];
    pc = instr->next;
    goto **pc;
  }

 l_ALOC: {
    instr_rr_t* instr = (instr_rr_t*)pc;
    regs[instr->ra] = (intptr_t)memory_allocate(regs[instr->rb]);
    pc = instr->next;
    goto **pc;
  }

 l_RCHR: {
    instr_r_t* instr = (instr_r_t*)pc;
    char input;
    fscanf(stdin, "%c", &input);
    regs[instr->r] = (intptr_t)input;
    pc = instr->next;
    goto **pc;
  }

 l_PCHR: {
    instr_r_t* instr = (instr_r_t*)pc;
    fprintf(stdout, "%c", (char)regs[instr->r]);
    fflush(stdout);
    pc = instr->next;
    goto **pc;
  }

 l_HALT: {
    return 0;
  }
}

static void check_available_memory(size_t size) {
  if (curr_addr + size > memory_get_end())
    fail("not enough memory for code");
}
